import os
from unittest.mock import Mock, patch

from bravado.exception import HTTPNotFound
from jsonschema.exceptions import RefResolutionError

from django.test import TestCase

from allianceauth import __url__ as aa_url
from allianceauth import __version__ as aa_version

from esi import __url__ as esi_url
from esi import __version__ as esi_version

from . import set_logger
from .esi_client_stub import EsiClientStub
from ..providers import (
    ObjectNotFound,
    Entity,
    AllianceMixin,
    FactionMixin,
    Character,
    Corporation,
    Alliance,
    ItemType,
    EveProvider,
    EveSwaggerProvider
)

MODULE_PATH = 'allianceauth.eveonline.providers'
SWAGGER_OLD_SPEC_PATH = os.path.join(os.path.dirname(
    os.path.abspath(__file__)), 'swagger_old.json'
)
set_logger(MODULE_PATH, __file__)


class TestObjectNotFound(TestCase):

    def test_str(self):
        x = ObjectNotFound(1001, 'Character')
        self.assertEqual(str(x), 'Character with ID 1001 not found.')


class TestEntity(TestCase):

    def test_str(self):
        x = Entity(1001, 'Bruce Wayne')
        self.assertEqual(str(x), 'Bruce Wayne')

        # bug - does not return a string
        """
        x = Entity(1001)
        self.assertEqual(str(x), '')

        x = Entity()
        self.assertEqual(str(x), '')
        """

    def test_repr(self):
        x = Entity(1001, 'Bruce Wayne')
        self.assertEqual(repr(x), '<Entity (1001): Bruce Wayne>')

        x = Entity(1001)
        self.assertEqual(repr(x), '<Entity (1001): None>')

        x = Entity()
        self.assertEqual(repr(x), '<Entity (None): None>')

    def test_bool(self):
        x = Entity(1001)
        self.assertTrue(bool(x))

        x = Entity()
        self.assertFalse(bool(x))

    def test_eq(self):
        x1 = Entity(1001)
        x2 = Entity(1001)
        y = Entity(1002)
        z1 = Entity()
        z2 = Entity()

        self.assertEqual(x1, x2)
        self.assertNotEqual(x1, y)
        self.assertNotEqual(x1, z1)
        self.assertEqual(z1, z2)

        # bug: missing _neq_ in Equity to compliment _eq_


class TestAllianceMixin(TestCase):
    @patch(MODULE_PATH + '.EveSwaggerProvider.get_alliance')
    def test_alliance_defined(self, mock_provider_get_alliance):
        my_alliance = Alliance(
            id=3001,
            name='Dummy Alliance',
            ticker='Dummy',
            corp_ids=[2001, 2002, 2003],
            executor_corp_id=2001
        )
        mock_provider_get_alliance.return_value = my_alliance

        x = AllianceMixin(alliance_id=3001)
        self.assertEqual(
            x.alliance,
            my_alliance
        )
        self.assertEqual(
            x.alliance,
            my_alliance
        )
        # should fetch alliance once only
        self.assertEqual(mock_provider_get_alliance.call_count, 1)

    @patch(MODULE_PATH + '.EveSwaggerProvider.get_alliance')
    def test_alliance_not_defined(self, mock_provider_get_alliance):
        mock_provider_get_alliance.return_value = None

        x = AllianceMixin()
        self.assertEqual(
            x.alliance,
            Entity(None, None)
        )
        self.assertEqual(mock_provider_get_alliance.call_count, 0)


class TestFactionMixin(TestCase):
    @patch(MODULE_PATH + '.EveSwaggerProvider.get_faction')
    def test_faction_defined(self, mock_provider_get_faction):
        my_faction = Entity(id=1337, name='Permabanned')
        mock_provider_get_faction.return_value = my_faction

        x = FactionMixin(faction_id=3001)
        self.assertEqual(
            x.faction,
            my_faction
        )
        self.assertEqual(
            x.faction,
            my_faction
        )
        # should fetch alliance once only
        self.assertEqual(mock_provider_get_faction.call_count, 1)

    @patch(MODULE_PATH + '.EveSwaggerProvider.get_alliance')
    def test_faction_not_defined(self, mock_provider_get_faction):
        mock_provider_get_faction.return_value = None

        x = FactionMixin()
        self.assertEqual(
            x.faction,
            Entity(None, None)
        )
        self.assertEqual(mock_provider_get_faction.call_count, 0)


class TestCorporation(TestCase):

    @patch(MODULE_PATH + '.EveSwaggerProvider.get_alliance')
    def test_alliance_defined(self, mock_provider_get_alliance):
        my_alliance = Alliance(
            id=3001,
            name='Dummy Alliance',
            ticker='Dummy',
            corp_ids=[2001, 2002, 2003],
            executor_corp_id=2001
        )
        mock_provider_get_alliance.return_value = my_alliance

        x = Corporation(alliance_id=3001)
        self.assertEqual(
            x.alliance,
            my_alliance
        )
        self.assertEqual(
            x.alliance,
            my_alliance
        )
        # should fetch alliance once only
        self.assertEqual(mock_provider_get_alliance.call_count, 1)

    @patch(MODULE_PATH + '.EveSwaggerProvider.get_alliance')
    def test_alliance_not_defined(self, mock_provider_get_alliance):
        mock_provider_get_alliance.return_value = None

        x = Corporation()
        self.assertEqual(
            x.alliance,
            Entity(None, None)
        )
        self.assertEqual(mock_provider_get_alliance.call_count, 0)

    @patch(MODULE_PATH + '.EveSwaggerProvider.get_character')
    def test_ceo(self, mock_provider_get_character):
        my_ceo = Character(
            id=1001,
            name='Bruce Wayne',
            corp_id=2001,
            alliance_id=3001
        )
        mock_provider_get_character.return_value = my_ceo

        # fetch from provider if not defined
        x = Corporation()
        self.assertEqual(
            x.ceo,
            my_ceo
        )

        # return existing if defined
        mock_provider_get_character.return_value = None
        self.assertEqual(
            x.ceo,
            my_ceo
        )
        self.assertEqual(mock_provider_get_character.call_count, 1)

        # bug in ceo(): will try to fetch character even if ceo_id is None

    @patch(MODULE_PATH + '.EveSwaggerProvider.get_faction')
    def test_faction_defined(self, mock_provider_get_faction):
        my_faction = Entity(id=1337, name='Permabanned')
        mock_provider_get_faction.return_value = my_faction

        # fetch from provider if not defined
        x = Corporation(faction_id=1337)
        self.assertEqual(x.faction, my_faction)

        # return existing if defined
        mock_provider_get_faction.return_value = None
        self.assertEqual(x.faction, my_faction)
        self.assertEqual(mock_provider_get_faction.call_count, 1)

    @patch(MODULE_PATH + '.EveSwaggerProvider.get_faction')
    def test_faction_undefined(self, mock_provider_get_faction):
        x = Corporation()
        self.assertEqual(x.faction, Entity())
        self.assertEqual(mock_provider_get_faction.call_count, 0)


class TestAlliance(TestCase):

    def setUp(self):
        self.my_alliance = Alliance(
            id=3001,
            name='Dummy Alliance',
            ticker='Dummy',
            corp_ids=[2001, 2002, 2003],
            executor_corp_id=2001,
            faction_id=1337
        )

    @staticmethod
    def _get_corp(corp_id):
        corps = {
            2001: Corporation(
                id=2001,
                name='Dummy Corp 1',
                alliance_id=3001
            ),
            2002: Corporation(
                id=2002,
                name='Dummy Corp 2',
                alliance_id=3001
            ),
            2003: Corporation(
                id=2003,
                name='Dummy Corp 3',
                alliance_id=3001
            ),
        }

        if corp_id:
            return corps[int(corp_id)]

    @patch(MODULE_PATH + '.EveSwaggerProvider.get_corp')
    def test_corp(self, mock_provider_get_corp):
        mock_provider_get_corp.side_effect = TestAlliance._get_corp

        # should fetch corp if not in the object
        self.assertEqual(
            self.my_alliance.corp(2001),
            TestAlliance._get_corp(2001)
        )
        # should fetch corp if not in the object
        self.assertEqual(
            self.my_alliance.corp(2002),
            TestAlliance._get_corp(2002)
        )
        # should return from the object if its there
        self.assertEqual(
            self.my_alliance.corp(2001),
            TestAlliance._get_corp(2001)
        )
        # should return from the object if its there
        self.assertEqual(
            self.my_alliance.corp(2002),
            TestAlliance._get_corp(2002)
        )
        # should be called once by used corp only
        self.assertEqual(mock_provider_get_corp.call_count, 2)

    @patch(MODULE_PATH + '.EveSwaggerProvider.get_corp')
    def test_corps(self, mock_provider_get_corp):
        mock_provider_get_corp.side_effect = TestAlliance._get_corp

        self.assertEqual(
            self.my_alliance.corps,
            [
                TestAlliance._get_corp(2001),
                TestAlliance._get_corp(2002),
                TestAlliance._get_corp(2003),
            ]
        )

    @patch(MODULE_PATH + '.EveSwaggerProvider.get_corp')
    def test_executor_corp(self, mock_provider_get_corp):
        mock_provider_get_corp.side_effect = TestAlliance._get_corp

        self.assertEqual(
            self.my_alliance.executor_corp,
            TestAlliance._get_corp(2001),
        )

        x = Alliance()
        self.assertEqual(
            x.executor_corp,
            Entity(None, None),
        )

    @patch(MODULE_PATH + '.EveSwaggerProvider.get_faction')
    def test_faction_defined(self, mock_provider_get_faction):
        my_faction = Entity(id=1337, name='Permabanned')
        mock_provider_get_faction.return_value = my_faction

        # fetch from provider if not defined
        self.assertEqual(self.my_alliance.faction, my_faction)

        # return existing if defined
        mock_provider_get_faction.return_value = None
        self.assertEqual(self.my_alliance.faction, my_faction)
        self.assertEqual(mock_provider_get_faction.call_count, 1)

    @patch(MODULE_PATH + '.EveSwaggerProvider.get_faction')
    def test_faction_undefined(self, mock_provider_get_faction):
        self.my_alliance.faction_id = None
        self.assertEqual(self.my_alliance.faction, Entity())
        self.assertEqual(mock_provider_get_faction.call_count, 0)


class TestCharacter(TestCase):

    def setUp(self):
        self.my_character = Character(
            id=1001,
            name='Bruce Wayne',
            corp_id=2001,
            alliance_id=3001,
            faction_id=1337,
        )

    @patch(MODULE_PATH + '.EveSwaggerProvider.get_corp')
    def test_corp(self, mock_provider_get_corp):
        my_corp = Corporation(
            id=2001,
            name='Dummy Corp 1'
        )
        mock_provider_get_corp.return_value = my_corp

        self.assertEqual(self.my_character.corp, my_corp)
        self.assertEqual(self.my_character.corp, my_corp)

        # should call the provider one time only
        self.assertEqual(mock_provider_get_corp.call_count, 1)

    @patch(MODULE_PATH + '.EveSwaggerProvider.get_alliance')
    @patch(MODULE_PATH + '.EveSwaggerProvider.get_corp')
    def test_alliance_has_one(
        self,
        mock_provider_get_corp,
        mock_provider_get_alliance,
    ):
        my_corp = Corporation(
            id=2001,
            name='Dummy Corp 1',
            alliance_id=3001
        )
        mock_provider_get_corp.return_value = my_corp
        my_alliance = Alliance(
            id=3001,
            name='Dummy Alliance 1',
            executor_corp_id=2001,
            corp_ids=[2001, 2002]
        )
        mock_provider_get_alliance.return_value = my_alliance

        self.assertEqual(self.my_character.alliance, my_alliance)
        self.assertEqual(self.my_character.alliance, my_alliance)

        # should call the provider one time only
        self.assertEqual(mock_provider_get_alliance.call_count, 1)

    @patch(MODULE_PATH + '.EveSwaggerProvider.get_alliance')
    def test_alliance_has_none(self, mock_provider_get_alliance):
        self.my_character.alliance_id = None
        self.assertEqual(self.my_character.alliance, Entity(None, None))
        self.assertEqual(mock_provider_get_alliance.call_count, 0)

    @patch(MODULE_PATH + '.EveSwaggerProvider.get_faction')
    def test_faction_defined(self, mock_provider_get_faction):
        my_faction = Entity(id=1337, name='Permabanned')
        mock_provider_get_faction.return_value = my_faction

        # fetch from provider if not defined
        self.assertEqual(self.my_character.faction, my_faction)

        # return existing if defined
        mock_provider_get_faction.return_value = None
        self.assertEqual(self.my_character.faction, my_faction)
        self.assertEqual(mock_provider_get_faction.call_count, 1)

    @patch(MODULE_PATH + '.EveSwaggerProvider.get_faction')
    def test_faction_undefined(self, mock_provider_get_faction):
        self.my_character.faction_id = None
        self.assertEqual(self.my_character.faction, Entity())
        self.assertEqual(mock_provider_get_faction.call_count, 0)


class TestItemType(TestCase):

    def test_init(self):
        x = ItemType(id=99, name='Dummy Item')
        self.assertIsInstance(x, ItemType)


class TestEveProvider(TestCase):

    def setUp(self):
        self.my_provider = EveProvider()

    def test_get_alliance(self):
        with self.assertRaises(NotImplementedError):
            self.my_provider.get_alliance(3001)

    def test_get_corp(self):
        with self.assertRaises(NotImplementedError):
            self.my_provider.get_corp(2001)

    def test_get_character(self):
        with self.assertRaises(NotImplementedError):
            self.my_provider.get_character(1001)

    # bug: should be calling NotImplementedError() not NotImplemented
    """
    def test_get_itemtype(self):
        with self.assertRaises(NotImplementedError):
            self.my_provider.get_itemtype(4001)
    """


class TestEveSwaggerProvider(TestCase):

    @staticmethod
    def esi_get_alliances_alliance_id(alliance_id):
        alliances = {
            3001: {
                'name': 'Dummy Alliance 1',
                'ticker': 'DA1',
                'executor_corporation_id': 2001
            },
            3002: {
                'name': 'Dummy Alliance 2',
                'ticker': 'DA2'
            }
        }
        mock_result = Mock()
        if alliance_id in alliances:
            mock_result.result.return_value = alliances[alliance_id]
            return mock_result
        else:
            raise HTTPNotFound(Mock())

    @staticmethod
    def esi_get_alliances_alliance_id_corporations(alliance_id):
        alliances = {
            3001: [2001, 2002, 2003],
            3002: [2004, 2005]
        }
        mock_result = Mock()
        if alliance_id in alliances:
            mock_result.result.return_value = alliances[alliance_id]
            return mock_result
        else:
            raise HTTPNotFound(Mock())

    @staticmethod
    def esi_get_corporations_corporation_id(corporation_id):
        corporations = {
            2001: {
                'name': 'Dummy Corp 1',
                'ticker': 'DC1',
                'ceo_id': 1001,
                'member_count': 42,
                'alliance_id': 3001
            },
            2002: {
                'name': 'Dummy Corp 2',
                'ticker': 'DC2',
                'ceo_id': 1011,
                'member_count': 5
            }
        }
        mock_result = Mock()
        if corporation_id in corporations:
            mock_result.result.return_value = corporations[corporation_id]
            return mock_result
        else:
            raise HTTPNotFound(Mock())

    @staticmethod
    def esi_get_characters_character_id(character_id):
        characters = {
            1001: {
                'name': 'Bruce Wayne',
                'corporation_id': 2001,
                'alliance_id': 3001
            },
            1002: {
                'name': 'Peter Parker',
                'corporation_id': 2101
            }
        }
        mock_result = Mock()
        if character_id in characters:
            mock_result.result.return_value = characters[character_id]
            return mock_result
        else:
            raise HTTPNotFound(Mock())

    @staticmethod
    def esi_post_characters_affiliation(characters):
        character_data = {
            1001: {
                'corporation_id': 2001,
                'alliance_id': 3001
            },
            1002: {
                'corporation_id': 2101
            }
        }
        mock_result = Mock()
        if isinstance(characters, list):
            characters_result = list()
            for character_id in characters:
                if character_id in character_data:
                    characters_result.append(character_data[character_id])
                else:
                    raise HTTPNotFound(Mock())
            mock_result.result.return_value = characters_result
            return mock_result
        else:
            raise TypeError()

    @staticmethod
    def esi_get_universe_types_type_id(type_id):
        types = {
            4001: {
                'name': 'Dummy Type 1'
            },
            4002: {
                'name': 'Dummy Type 2'
            }
        }
        mock_result = Mock()
        if type_id in types:
            mock_result.result.return_value = types[type_id]
            return mock_result
        else:
            raise HTTPNotFound(Mock())

    @patch(MODULE_PATH + '.esi_client_factory')
    def test_str(self, mock_esi_client_factory):
        my_provider = EveSwaggerProvider()
        self.assertEqual(str(my_provider), 'esi')

    @patch(MODULE_PATH + '.esi_client_factory')
    def test_get_alliance(self, mock_esi_client_factory):
        mock_esi_client_factory.return_value \
            .Alliance.get_alliances_alliance_id \
            = TestEveSwaggerProvider.esi_get_alliances_alliance_id
        mock_esi_client_factory.return_value \
            .Alliance.get_alliances_alliance_id_corporations \
            = TestEveSwaggerProvider.esi_get_alliances_alliance_id_corporations

        my_provider = EveSwaggerProvider()

        # fully defined alliance
        my_alliance = my_provider.get_alliance(3001)
        self.assertEqual(my_alliance.id, 3001)
        self.assertEqual(my_alliance.name, 'Dummy Alliance 1')
        self.assertEqual(my_alliance.ticker, 'DA1')
        self.assertListEqual(my_alliance.corp_ids, [2001, 2002, 2003])
        self.assertEqual(my_alliance.executor_corp_id, 2001)

        # alliance missing executor_corporation_id
        my_alliance = my_provider.get_alliance(3002)
        self.assertEqual(my_alliance.id, 3002)
        self.assertEqual(my_alliance.executor_corp_id, None)

        # alliance not found
        with self.assertRaises(ObjectNotFound):
            my_provider.get_alliance(3999)

    @patch(MODULE_PATH + '.esi_client_factory')
    def test_get_corp(self, mock_esi_client_factory):
        mock_esi_client_factory.return_value \
            .Corporation.get_corporations_corporation_id \
            = TestEveSwaggerProvider.esi_get_corporations_corporation_id

        my_provider = EveSwaggerProvider()

        # corporation with alliance
        my_corp = my_provider.get_corp(2001)
        self.assertEqual(my_corp.id, 2001)
        self.assertEqual(my_corp.name, 'Dummy Corp 1')
        self.assertEqual(my_corp.ticker, 'DC1')
        self.assertEqual(my_corp.ceo_id, 1001)
        self.assertEqual(my_corp.members, 42)
        self.assertEqual(my_corp.alliance_id, 3001)

        # corporation wo/ alliance
        my_corp = my_provider.get_corp(2002)
        self.assertEqual(my_corp.id, 2002)
        self.assertEqual(my_corp.alliance_id, None)

        # corporation not found
        with self.assertRaises(ObjectNotFound):
            my_provider.get_corp(2999)

    @patch(MODULE_PATH + '.esi_client_factory')
    def test_get_character(self, mock_esi_client_factory):
        mock_esi_client_factory.return_value = EsiClientStub()
        my_provider = EveSwaggerProvider()

        # character with alliance
        my_character = my_provider.get_character(1001)
        self.assertEqual(my_character.id, 1001)
        self.assertEqual(my_character.name, 'Bruce Wayne')
        self.assertEqual(my_character.corp_id, 2001)
        self.assertEqual(my_character.alliance_id, 3001)

        # character wo/ alliance
        my_character = my_provider.get_character(1011)
        self.assertEqual(my_character.id, 1011)
        self.assertEqual(my_character.alliance_id, None)

        # character not found
        with self.assertRaises(ObjectNotFound):
            my_provider.get_character(1999)

    @patch(MODULE_PATH + '.esi_client_factory')
    def test_get_itemtype(self, mock_esi_client_factory):
        mock_esi_client_factory.return_value \
            .Universe.get_universe_types_type_id \
            = TestEveSwaggerProvider.esi_get_universe_types_type_id

        my_provider = EveSwaggerProvider()

        # type exists
        my_type = my_provider.get_itemtype(4001)
        self.assertEqual(my_type.id, 4001)
        self.assertEqual(my_type.name, 'Dummy Type 1')

        # type not found
        with self.assertRaises(ObjectNotFound):
            my_provider.get_itemtype(4999)

    @patch(MODULE_PATH + '.settings.DEBUG', False)
    @patch(MODULE_PATH + '.esi_client_factory')
    def test_create_client_on_normal_startup(self, mock_esi_client_factory):
        my_provider = EveSwaggerProvider()
        self.assertTrue(mock_esi_client_factory.called)
        self.assertIsNotNone(my_provider._client)

    @patch(MODULE_PATH + '.SWAGGER_SPEC_PATH', SWAGGER_OLD_SPEC_PATH)
    @patch(MODULE_PATH + '.settings.DEBUG', False)
    @patch('socket.socket')
    def test_create_client_on_normal_startup_w_old_swagger_spec(
            self, mock_socket
    ):
        mock_socket.side_effect = Exception('Network blocked for testing')
        my_provider = EveSwaggerProvider()
        self.assertIsNone(my_provider._client)

    @patch(MODULE_PATH + '.settings.DEBUG', True)
    @patch(MODULE_PATH + '.esi_client_factory')
    def test_dont_create_client_on_debug_startup(self, mock_esi_client_factory):
        my_provider = EveSwaggerProvider()
        self.assertFalse(mock_esi_client_factory.called)
        self.assertIsNone(my_provider._client)

    @patch(MODULE_PATH + '.settings.DEBUG', False)
    @patch(MODULE_PATH + '.esi_client_factory')
    def test_dont_create_client_if_client_creation_fails_on_normal_startup(
            self, mock_esi_client_factory
    ):
        mock_esi_client_factory.side_effect = RefResolutionError(cause='Test')
        my_provider = EveSwaggerProvider()
        self.assertTrue(mock_esi_client_factory.called)
        self.assertIsNone(my_provider._client)

    @patch(MODULE_PATH + '.settings.DEBUG', True)
    @patch(MODULE_PATH + '.esi_client_factory')
    def test_client_loads_on_demand(
            self, mock_esi_client_factory
    ):
        mock_esi_client_factory.return_value = 'my_client'
        my_provider = EveSwaggerProvider()
        self.assertFalse(mock_esi_client_factory.called)
        self.assertIsNone(my_provider._client)
        my_client = my_provider.client
        self.assertTrue(mock_esi_client_factory.called)
        self.assertIsNotNone(my_provider._client)
        self.assertEqual(my_client, 'my_client')

    def test_user_agent_header(self):
        my_provider = EveSwaggerProvider()
        my_client = my_provider.client
        operation = my_client.Universe.get_universe_factions()
        self.assertEqual(
            operation.future.request.headers['User-Agent'],
            f'AllianceAuth/{aa_version} (dummy@example.net; +{aa_url}) Django-ESI/{esi_version} (+{esi_url})'
        )



from typing import List, Optional, Union


class ThemeHook:
    """
    Theme hook for injecting a Bootstrap 5 Theme and associated JS into alliance auth.
    These can be local or CDN delivered.
    """

    def __init__(self,
        name: str,
        description: str,
        css: List[dict],
        js: List[dict],
        css_template: Optional[str] = None,
        js_template: Optional[str] = None,
        js_type: Optional[str] = None,
        html_tags: Optional[Union[dict, str]] = None,
        header_padding: Optional[str] = "4em"
    ):
        """
        :param name: Theme python name
        :type name: str
        :param description: Theme verbose name
        :type description: str
        :param css: CSS paths to load
        :type css: List[dict]
        :param js: JS paths to load
        :type js: List[dict]
        :param css_template: _description_, defaults to None
        :type css_template: Optional[str], optional
        :param js_template: _description_, defaults to None
        :type js_template: Optional[str], optional
        :param js_type: The type of the JS (e.g.: 'module'), defaults to None
        :type js_type: Optional[str], optional
        :param html_tags: Attributes added to the `<html>` tag, defaults to None
        :type html_tags: Optional[dict|str], optional
        :param header_padding: Top padding, defaults to "4em"
        :type header_padding: Optional[str], optional
        """

        self.name = name
        self.description = description

        # Direct from CDN
        self.css = css
        self.js = js

        # Load a django template with static file definitions
        self.css_template = css_template
        self.js_template = js_template

        # Define the JS type (e.g.: 'module')
        self.js_type = js_type

        self.html_tags = (
            " ".join([f"{key}={value}" for key, value in html_tags.items()])
            if isinstance(html_tags, dict)
            else html_tags
        )
        self.header_padding = header_padding

    def get_name(self):
        return f"{self.__class__.__module__}.{self.__class__.__name__}"



from unittest.mock import Mock, patch

from requests.exceptions import HTTPError

from allianceauth.tests.auth_utils import AuthUtils
from allianceauth.utils.testing import NoSocketsTestCase

from ..discord_client import DiscordApiBackoff, RolesSet
from ..discord_client.tests.factories import (
    TEST_USER_ID,
    TEST_USER_NAME,
    create_guild_member,
    create_role,
)
from ..discord_client.tests.factories import create_user as create_guild_user
from ..models import DiscordUser
from ..utils import set_logger_to_file
from . import MODULE_PATH, TEST_MAIN_ID, TEST_MAIN_NAME
from .factories import create_discord_user, create_user

logger = set_logger_to_file(MODULE_PATH + '.models', __file__)


class TestBasicsAndHelpers(NoSocketsTestCase):

    def test_str(self):
        user = AuthUtils.create_user(TEST_USER_NAME)
        discord_user = DiscordUser.objects.create(user=user, uid=TEST_USER_ID)
        expected = 'Peter Parker - 198765432012345678'
        self.assertEqual(str(discord_user), expected)

    def test_repr(self):
        user = AuthUtils.create_user(TEST_USER_NAME)
        discord_user = DiscordUser.objects.create(user=user, uid=TEST_USER_ID)
        expected = 'DiscordUser(user=\'Peter Parker\', uid=198765432012345678)'
        self.assertEqual(repr(discord_user), expected)


@patch(MODULE_PATH + '.models.default_bot_client', spec=True)
class TestUpdateNick(NoSocketsTestCase):

    def setUp(self):
        self.user = AuthUtils.create_user(TEST_USER_NAME)
        self.discord_user = DiscordUser.objects.create(
            user=self.user, uid=TEST_USER_ID
        )

    def test_can_update(self, mock_default_bot_client):
        # given
        AuthUtils.add_main_character_2(
            self.user, TEST_MAIN_NAME, TEST_MAIN_ID, disconnect_signals=True
        )
        mock_default_bot_client.modify_guild_member.return_value = True
        # when
        result = self.discord_user.update_nickname()
        # then
        self.assertTrue(result)
        self.assertTrue(mock_default_bot_client.modify_guild_member.called)

    def test_dont_update_if_user_has_no_main(self, mock_default_bot_client):
        # given
        mock_default_bot_client.modify_guild_member.return_value = False
        # when
        result = self.discord_user.update_nickname()
        # then
        self.assertFalse(result)
        self.assertFalse(mock_default_bot_client.modify_guild_member.called)

    def test_return_none_if_user_no_longer_a_member(self, mock_default_bot_client):
        # given
        AuthUtils.add_main_character_2(
            self.user, TEST_MAIN_NAME, TEST_MAIN_ID, disconnect_signals=True
        )
        mock_default_bot_client.modify_guild_member.return_value = None
        # when
        result = self.discord_user.update_nickname()
        # then
        self.assertIsNone(result)
        self.assertTrue(mock_default_bot_client.modify_guild_member.called)

    def test_return_false_if_api_returns_false(self, mock_default_bot_client):
        # given
        AuthUtils.add_main_character_2(
            self.user, TEST_MAIN_NAME, TEST_MAIN_ID, disconnect_signals=True
        )
        mock_default_bot_client.modify_guild_member.return_value = False
        # when
        result = self.discord_user.update_nickname()
        # then
        self.assertFalse(result)
        self.assertTrue(mock_default_bot_client.modify_guild_member.called)


@patch(MODULE_PATH + '.models.default_bot_client.guild_member', spec=True)
class TestUpdateUsername(NoSocketsTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = create_user()

    def test_can_update(self, mock_guild_member):
        # given
        discord_user = create_discord_user(user=self.user)
        new_username = 'New name'
        new_discriminator = '9876'
        guild_user = create_guild_user(
            username='New name', discriminator=new_discriminator
        )
        mock_guild_member.return_value = create_guild_member(user=guild_user)
        # when
        result = discord_user.update_username()
        # then
        self.assertTrue(result)
        self.assertTrue(mock_guild_member.called)
        discord_user.refresh_from_db()
        self.assertEqual(discord_user.username, new_username)
        self.assertEqual(discord_user.discriminator, new_discriminator)

    def test_return_none_if_user_no_longer_a_member(self, mock_guild_member):
        # given
        discord_user = create_discord_user(user=self.user)
        mock_guild_member.return_value = None
        # when
        result = discord_user.update_username()
        # then
        self.assertIsNone(result)
        self.assertTrue(mock_guild_member.called)


@patch(MODULE_PATH + '.models.notify', spec=True)
@patch(MODULE_PATH + '.models.create_bot_client', spec=True)
class TestDeleteUser(NoSocketsTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = AuthUtils.create_user(TEST_USER_NAME)

    def setUp(self):
        self.discord_user = DiscordUser.objects.create(
            user=self.user, uid=TEST_USER_ID
        )

    def test_can_delete_user(self, mock_create_bot_client, mock_notify):
        # given
        mock_create_bot_client.return_value.remove_guild_member.return_value = True
        # when
        result = self.discord_user.delete_user()
        # then
        self.assertTrue(result)
        self.assertFalse(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        self.assertTrue(mock_create_bot_client.return_value.remove_guild_member.called)
        self.assertFalse(mock_notify.called)

    def test_can_delete_user_and_notify_user(self, mock_create_bot_client, mock_notify):
        # given
        mock_create_bot_client.return_value.remove_guild_member.return_value = True
        # when
        result = self.discord_user.delete_user(notify_user=True)
        # then
        self.assertTrue(result)
        self.assertTrue(mock_notify.called)

    def test_can_delete_user_when_member_is_unknown(
        self, mock_create_bot_client, mock_notify
    ):
        # given
        mock_create_bot_client.return_value.remove_guild_member.return_value = None
        # when
        result = self.discord_user.delete_user()
        # then
        self.assertTrue(result)
        self.assertFalse(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        self.assertTrue(mock_create_bot_client.return_value.remove_guild_member.called)
        self.assertFalse(mock_notify.called)

    def test_return_false_when_api_fails(self, mock_create_bot_client, mock_notify):
        # given
        mock_create_bot_client.return_value.remove_guild_member.return_value = False
        # when
        result = self.discord_user.delete_user()
        # then
        self.assertFalse(result)

    def test_dont_notify_if_user_was_already_deleted_and_return_none(
        self, mock_create_bot_client, mock_notify
    ):
        # given
        mock_create_bot_client.return_value.remove_guild_member.return_value = None
        DiscordUser.objects.get(pk=self.discord_user.pk).delete()
        # when
        result = self.discord_user.delete_user()
        # then
        self.assertIsNone(result)
        self.assertFalse(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        self.assertTrue(mock_create_bot_client.return_value.remove_guild_member.called)
        self.assertFalse(mock_notify.called)

    def test_raise_exception_on_api_backoff(
        self, mock_create_bot_client, mock_notify
    ):
        # given
        mock_create_bot_client.return_value.remove_guild_member.side_effect = \
            DiscordApiBackoff(999)
        # when/then
        with self.assertRaises(DiscordApiBackoff):
            self.discord_user.delete_user()

    def test_return_false_on_api_backoff_and_exception_handling_on(
        self, mock_create_bot_client, mock_notify
    ):
        # given
        mock_create_bot_client.return_value.remove_guild_member.side_effect = \
            DiscordApiBackoff(999)
        # when
        result = self.discord_user.delete_user(handle_api_exceptions=True)
        # then
        self.assertFalse(result)

    def test_raise_exception_on_http_error(
        self, mock_create_bot_client, mock_notify
    ):
        # given
        mock_exception = HTTPError('error')
        mock_exception.response = Mock()
        mock_exception.response.status_code = 500
        mock_create_bot_client.return_value.remove_guild_member.side_effect = \
            mock_exception
        # when/then
        with self.assertRaises(HTTPError):
            self.discord_user.delete_user()

    def test_return_false_on_http_error_and_exception_handling_on(
        self, mock_create_bot_client, mock_notify
    ):
        # given
        mock_exception = HTTPError('error')
        mock_exception.response = Mock()
        mock_exception.response.status_code = 500
        mock_create_bot_client.return_value.remove_guild_member.side_effect = \
            mock_exception
        # when
        result = self.discord_user.delete_user(handle_api_exceptions=True)
        # then
        self.assertFalse(result)


@patch(MODULE_PATH + '.models.default_bot_client', spec=True)
@patch(MODULE_PATH + '.models.calculate_roles_for_user', spec=True)
class TestUpdateGroups(NoSocketsTestCase):

    def setUp(self):
        user = AuthUtils.create_user(TEST_USER_NAME)
        self.discord_user = DiscordUser.objects.create(user=user, uid=TEST_USER_ID)

    def test_should_update_when_roles_have_changed(
        self, mock_calculate_roles_for_user, mock_client
    ):
        # given
        mock_calculate_roles_for_user.return_value = RolesSet([create_role()]), True
        mock_client.modify_guild_member.return_value = True
        # when
        result = self.discord_user.update_groups()
        # then
        self.assertTrue(result)
        self.assertTrue(mock_client.modify_guild_member.called)

    def test_should_not_update_when_roles_have_not_changed(
        self, mock_calculate_roles_for_user, mock_client
    ):
        # given
        mock_calculate_roles_for_user.return_value = RolesSet([create_role()]), False
        mock_client.modify_guild_member.return_value = True
        # when
        result = self.discord_user.update_groups()
        # then
        self.assertTrue(result)
        self.assertFalse(mock_client.modify_guild_member.called)

    def test_should_not_update_when_user_not_guild_member(
        self, mock_calculate_roles_for_user, mock_client
    ):
        # given
        mock_calculate_roles_for_user.return_value = RolesSet([create_role()]), None
        mock_client.modify_guild_member.return_value = True
        # when
        result = self.discord_user.update_groups()
        # then
        self.assertIsNone(result)
        self.assertFalse(mock_client.modify_guild_member.called)

    def test_should_return_false_when_update_failed(
        self, mock_calculate_roles_for_user, mock_client
    ):
        # given
        mock_calculate_roles_for_user.return_value = RolesSet([create_role()]), True
        mock_client.modify_guild_member.return_value = False
        # when
        result = self.discord_user.update_groups()
        # then
        self.assertFalse(result)
        self.assertTrue(mock_client.modify_guild_member.called)



"""
Template tags for language mapping
"""

from django.conf import settings
from django.template.defaulttags import register
from django.templatetags.static import static


@register.simple_tag
def get_datatable_language_code(language: str) -> str:
    """
    Get the correct language code for DataTables

    :param language: Django's language code
    :type language: str
    :return: Mapped language code
    :rtype: str
    """

    mapped_language = (
        settings.LANGUAGE_MAPPING["DataTables"].get(language)
        if language != "en"
        else ""
    )

    return mapped_language


@register.simple_tag
def get_momentjs_language_code(language: str) -> str:
    """
    Get the correct language code for Moment.JS

    :param language: Django's language code
    :type language: str
    :return: Mapped language code
    :rtype: str
    """

    mapped_language = (
        settings.LANGUAGE_MAPPING["MomentJS"].get(language) if language != "en" else ""
    )

    return mapped_language


@register.simple_tag
def get_datatables_language_static(language: str) -> str:
    """
    Get the correct language code URL for DataTables

    :param language: Django's language code
    :type language: str
    :return: Mapped language code
    :rtype: str
    """

    mapped_language = get_datatable_language_code(language)
    static_url = (
        static(
            path=f"allianceauth/libs/DataTables/Plugins/2.2.1/i18n/{mapped_language}.json"
        )
        if mapped_language
        else ""
    )

    return static_url

@register.simple_tag
def get_relative_datatables_language_path(language: str) -> str:
    """
    Get the correct language code URL for DataTables (relative path to the static folder)

    :param language: Django's language code
    :type language: str
    :return: Mapped language code
    :rtype: str
    """

    mapped_language = get_datatable_language_code(language)
    static_url = (
        f"allianceauth/libs/DataTables/Plugins/2.2.1/i18n/{mapped_language}.json"
        if mapped_language
        else ""
    )

    return static_url


@register.simple_tag
def get_momentjs_language_static(language: str) -> str:
    """
    Get the correct language code URL for Moment.JS

    :param language: Django's language code
    :type language: str
    :return: Mapped language code
    :rtype: str
    """

    mapped_language = get_momentjs_language_code(language)

    static_url = (
        static(path=f"allianceauth/libs/moment.js/2.29.4/locale/{mapped_language}.js")
        if mapped_language
        else ""
    )

    return static_url

@register.simple_tag
def get_relative_momentjs_language_path(language: str) -> str:
    """
    Get the correct language code URL for Moment.JS (relative path to the static folder)

    :param language: Django's language code
    :type language: str
    :return: Mapped language code path
    :rtype: str
    """

    mapped_language = get_momentjs_language_code(language)

    static_url = (
        f"allianceauth/libs/moment.js/2.29.4/locale/{mapped_language}.js"
        if mapped_language
        else ""
    )

    return static_url



import functools

from django import forms
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.contrib.auth.models import Group, User
from django.core.exceptions import ValidationError
from django.db.models.functions import Lower
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from .models import ReservedGroupName


class GroupAdminForm(forms.ModelForm):
    users = forms.ModelMultipleChoiceField(
        queryset=User.objects.order_by(Lower('username')),
        required=False,
        widget=FilteredSelectMultiple(verbose_name=_("Users"), is_stacked=False),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.instance and self.instance.pk:
            self.fields["users"].initial = self.instance.user_set.all()

    def save(self, commit=True):
        group: Group = super().save(commit=False)

        if commit:
            group.save()

        users = self.cleaned_data["users"]
        if group.pk:
            self._save_m2m_and_users(group, users)
        else:
            self.save_m2m = functools.partial(
                self._save_m2m_and_users, group=group, users=users
            )

        return group

    def _save_m2m_and_users(self, group, users):
        """Save m2m relations incl. users."""
        group.user_set.set(users)
        self._save_m2m()

    def clean_name(self):
        my_name = self.cleaned_data['name']
        if ReservedGroupName.objects.filter(name__iexact=my_name).exists():
            raise ValidationError(
                _("This name has been reserved and can not be used for groups."),
                code='reserved_name'
            )
        return my_name


class ReservedGroupNameAdminForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['created_by'].initial = self.current_user.username
        self.fields['created_at'].initial = _("(auto)")

    created_by = forms.CharField(disabled=True)
    created_at = forms.CharField(disabled=True)

    def clean_name(self):
        my_name = self.cleaned_data['name'].lower()
        if Group.objects.filter(name__iexact=my_name).exists():
            raise ValidationError(
                _("There already exists a group with that name."), code='already_exists'
            )
        return my_name

    def clean_created_at(self):
        return now()



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class NotificationsConfig(AppConfig):
    name = 'allianceauth.notifications'
    label = 'notifications'
    verbose_name = _('Notifications')



from http import HTTPStatus

from django.test import TestCase

from allianceauth.menu.core.smart_sync import reset_menu_items_sync
from allianceauth.menu.tests.factories import (
    create_folder_menu_item,
    create_link_menu_item,
    create_user,
)
from allianceauth.menu.tests.utils import extract_links


class TestDefaultDashboardWithSideMenu(TestCase):
    def test_should_show_all_types_of_menu_entries(self):
        # given
        user = create_user(permissions=["auth.group_management"])
        self.client.force_login(user)
        create_link_menu_item(text="Alpha", url="http://www.example.com/alpha")
        folder = create_folder_menu_item(text="Folder")
        create_link_menu_item(
            text="Bravo", url="http://www.example.com/bravo", parent=folder
        )
        reset_menu_items_sync()  # this simulates startup

        # when
        response = self.client.get("/dashboard/")

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        links = extract_links(response)
        # open_page_in_browser(response)
        self.assertEqual(links["/dashboard/"], "Dashboard")
        self.assertEqual(links["/groups/"], "Groups")
        self.assertEqual(links["/groupmanagement/requests/"], "Group Management")
        self.assertEqual(links["http://www.example.com/alpha"], "Alpha")
        self.assertEqual(links["http://www.example.com/bravo"], "Bravo")

    def test_should_not_show_menu_entry_when_user_has_no_permission(self):
        # given
        user = create_user()
        self.client.force_login(user)
        reset_menu_items_sync()

        # when
        response = self.client.get("/dashboard/")

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        links = extract_links(response)
        self.assertEqual(links["/dashboard/"], "Dashboard")
        self.assertEqual(links["/groups/"], "Groups")
        self.assertNotIn("/groupmanagement/requests/", links)

    def test_should_not_show_menu_entry_when_hidden(self):
        # given
        user = create_user()
        self.client.force_login(user)
        create_link_menu_item(text="Alpha", url="http://www.example.com/")
        reset_menu_items_sync()

        # when
        response = self.client.get("/dashboard/")

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        links = extract_links(response)
        self.assertEqual(links["/dashboard/"], "Dashboard")
        self.assertEqual(links["/groups/"], "Groups")
        self.assertNotIn("http://www.example.com/alpha", links)


class TestBS3DashboardWithSideMenu(TestCase):
    def test_should_not_show_group_management_when_user_has_no_permission(self):
        # given
        user = create_user()
        self.client.force_login(user)

        # when
        response = self.client.get("/dashboard_bs3/")

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        links = extract_links(response)
        self.assertEqual(links["/dashboard/"], "Dashboard")
        self.assertEqual(links["/groups/"], "Groups")
        self.assertNotIn("/groupmanagement/requests/", links)

    def test_should_show_group_management_when_user_has_permission(self):
        # given
        user = create_user(permissions=["auth.group_management"])
        self.client.force_login(user)

        # when
        response = self.client.get("/dashboard_bs3/")

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        links = extract_links(response)
        self.assertEqual(links["/dashboard/"], "Dashboard")
        self.assertEqual(links["/groups/"], "Groups")
        self.assertEqual(links["/groupmanagement/requests/"], "Group Management")



from django.db import models
from django.contrib.auth.models import User, Group
from allianceauth.authentication.models import State


class Teamspeak3User(models.Model):
    user = models.OneToOneField(User,
                                primary_key=True,
                                on_delete=models.CASCADE,
                                related_name='teamspeak3')
    uid = models.CharField(max_length=254)
    perm_key = models.CharField(max_length=254)

    def __str__(self):
        return self.uid

    class Meta:
        permissions = (
            ("access_teamspeak3", "Can access the Teamspeak3 service"),
        )


class TSgroup(models.Model):
    ts_group_id = models.IntegerField(primary_key=True)
    ts_group_name = models.CharField(max_length=30)

    class Meta:
        verbose_name = 'TS Group'

    def __str__(self):
        return self.ts_group_name


class AuthTS(models.Model):
    auth_group = models.ForeignKey(Group, on_delete=models.CASCADE)
    ts_group = models.ManyToManyField(TSgroup)

    class Meta:
        verbose_name = 'Auth / TS Group'

    def __str__(self):
        return self.auth_group.name


class UserTSgroup(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    ts_group = models.ManyToManyField(TSgroup)

    class Meta:
        verbose_name = 'User TS Group'

    def __str__(self):
        return self.user.name


class StateGroup(models.Model):
    state = models.ForeignKey(State, on_delete=models.CASCADE)
    ts_group = models.ForeignKey(TSgroup, on_delete=models.CASCADE)



from django.test import RequestFactory, TestCase

from allianceauth.menu.hooks import MenuItemHook

from .factories import create_menu_item_hook


class TestMenuItemHook(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.factory = RequestFactory()

    def test_should_create_obj_with_minimal_params(self):
        # when
        obj = MenuItemHook("text", "classes", "url-name")

        # then
        self.assertEqual(obj.text, "text")
        self.assertEqual(obj.classes, "classes")
        self.assertEqual(obj.url_name, "url-name")
        self.assertEqual(obj.template, "public/menuitem.html")
        self.assertEqual(obj.order, 9999)
        self.assertListEqual(obj.navactive, ["url-name"])
        self.assertIsNone(obj.count)

    def test_should_create_obj_with_full_params_1(self):
        # when
        obj = MenuItemHook("text", "classes", "url-name", 5, ["navactive"])

        # then
        self.assertEqual(obj.text, "text")
        self.assertEqual(obj.classes, "classes")
        self.assertEqual(obj.url_name, "url-name")
        self.assertEqual(obj.template, "public/menuitem.html")
        self.assertEqual(obj.order, 5)
        self.assertListEqual(obj.navactive, ["navactive", "url-name"])
        self.assertIsNone(obj.count)

    def test_should_create_obj_with_full_params_2(self):
        # when
        obj = MenuItemHook(
            text="text",
            classes="classes",
            url_name="url-name",
            order=5,
            navactive=["navactive"],
        )

        # then
        self.assertEqual(obj.text, "text")
        self.assertEqual(obj.classes, "classes")
        self.assertEqual(obj.url_name, "url-name")
        self.assertEqual(obj.template, "public/menuitem.html")
        self.assertEqual(obj.order, 5)
        self.assertListEqual(obj.navactive, ["navactive", "url-name"])
        self.assertIsNone(obj.count)

    def test_should_render_menu_item(self):
        # given
        request = self.factory.get("/")
        hook = create_menu_item_hook(text="Alpha")

        # when
        result = hook.render(request)

        # then
        self.assertIn("Alpha", result)

    def test_str(self):
        # given
        hook = create_menu_item_hook(text="Alpha")

        # when/then
        self.assertEqual(str(hook), "Alpha")

    def test_repr(self):
        # given
        hook = create_menu_item_hook(text="Alpha")

        # when/then
        self.assertIn("Alpha", repr(hook))
